﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Threading.Tasks;

namespace SqliteUsingSQLiteWinRT.ViewModels
{
  public class ProjectsViewModel : TableViewModelBase<ProjectViewModel, int>
  {
    private ProjectsViewModel(int customerId)
    {
      CustomerId = customerId;
    }

    private ProjectsViewModel()
    {
      CustomerId = -1;
    }

    static Dictionary<int, ProjectsViewModel> instances = new Dictionary<int, ProjectsViewModel>();
    static ProjectsViewModel defaultInstance;

    public override DateTime Timestamp
    {
      get
      {
        if (this == defaultInstance || defaultInstance == null)
          return base.Timestamp;

        return defaultInstance.Timestamp;
      }
      protected set
      {
        if (this == defaultInstance || defaultInstance == null)
          base.Timestamp = value;
        else
          defaultInstance.Timestamp = value;
      }
    }

    public static ProjectsViewModel GetForCustomerId(int customerId)
    {
      lock (typeof(ProjectsViewModel))
      {
        if (instances.ContainsKey(customerId) != true)
          instances[customerId] = new ProjectsViewModel(customerId);
      }

      return instances[customerId];
    }

    public static ProjectsViewModel GetDefault()
    {
      lock (typeof(ProjectsViewModel))
      {
        if (defaultInstance == null)
          defaultInstance = new ProjectsViewModel();
      }

      return defaultInstance;
    }

    public int CustomerId { get; private set; }

    protected override string GetSelectAllSql()
    {
      if (CustomerId < 0)
        return @"SELECT Id, CustomerId, Name, Description, DueDate 
                           FROM Project
                           ORDER BY DueDate";
      else
        return @"SELECT Id, CustomerId, Name, Description, DueDate 
                           FROM Project
                           WHERE CustomerId = ?
                           ORDER BY DueDate";
    }

    protected override void FillSelectAllStatement(SQLiteWinRT.Statement statement)
    {
      if (CustomerId < 0)
        return;

      statement.BindIntParameterAt(1, CustomerId);
    }

    protected override ProjectViewModel CreateItem(SQLiteWinRT.Statement statement)
    {
      ProjectViewModel project;

      // Illustrate how you can use the Columns property if it has been enabled
      // The base class enables it for SelectItem, not SelectAlls
      // As you can see, it isn't super useful (all columns are text) and it adds
      // complexity to the data enumeration, so the feature is off by default
      var columns = statement.Columns;
      if (columns != null)
      {
        project = new ProjectViewModel(
          int.Parse(columns["Id"]),
          int.Parse(columns["CustomerId"]),
          columns["Name"],
          columns["Description"],
          DateTime.Parse(columns["DueDate"])
        );
      }

      // More typical way of accessing data
      else
      {
        project = new ProjectViewModel(
          statement.GetIntAt(0),
          statement.GetIntAt(1),
          statement.GetTextAt(2),
          statement.GetTextAt(3),
          DateTime.Parse(statement.GetTextAt(4))
        );
      }

      Debug.WriteLine("Selected Project name:" + project.Name);
      return project;
    }

    protected override string GetSelectItemSql()
    {
      return @"SELECT Id, CustomerId, Name, Description, DueDate 
                           FROM Project
                           WHERE Id = ?";
    }

    protected override void FillSelectItemStatement(SQLiteWinRT.Statement statement, int key)
    {
      statement.BindIntParameterAt(1, key);
    }

    protected override string GetInsertItemSql()
    {
      return "INSERT INTO Project (CustomerId, Name, Description, DueDate) VALUES (?, ?, ?, ?)";
    }

    protected override void FillInsertStatement(SQLiteWinRT.Statement statement, ProjectViewModel item)
    {
      statement.BindIntParameterAt(1, item.CustomerId);
      statement.BindTextParameterAt(2, item.Name);
      statement.BindTextParameterAt(3, item.Description);
      statement.BindTextParameterAt(4, item.DueDate.ToString("yyyy-MM-dd HH:mm:ss"));
    }

    protected override string GetUpdateItemSql()
    {
      return "UPDATE Project SET CustomerId = ?, Name = ?, Description = ?, DueDate = ? WHERE Id = ?";
    }

    protected override void FillUpdateStatement(SQLiteWinRT.Statement statement, int key, ProjectViewModel item)
    {
      FillInsertStatement(statement, item);
      statement.BindIntParameterAt(5, key);
    }

    protected override string GetDeleteItemSql()
    {
      return "DELETE FROM Project WHERE Id = ?";
    }

    protected override void FillDeleteItemStatement(SQLiteWinRT.Statement statement, int key)
    {
      statement.BindIntParameterAt(1, key);
    }
  }
}
