﻿using Microsoft.Phone.Controls;
using SqliteUsingSQLiteWinRT.ViewModels;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading.Tasks;
using System.Windows.Controls;
using System.Windows.Navigation;

namespace SqliteUsingSQLiteWinRT.Views
{
  public partial class MainPage : PhoneApplicationPage
  {
    CustomersViewModel viewModel;
    DateTime timestamp = DateTime.MinValue;

    public MainPage()
    {
      InitializeComponent();
    }

    protected override void OnNavigatedTo(NavigationEventArgs e)
    {
      if (viewModel != null && timestamp >= viewModel.Timestamp)
        return;

      LoadData();

      timestamp = DateTime.Now;
    }

    private async void LoadData()
    {
      viewModel = CustomersViewModel.GetDefault();
      CustomersList.ItemsSource = await viewModel.GetAllItems();
    }

    private void AddCustomerAppBarButton_Click(object sender, EventArgs e)
    {
      NavigationService.Navigate(new Uri("/Views/CustomerEditView.xaml?NEW=true", UriKind.Relative));
    }

    private void CustomersList_SelectionChanged(object sender, SelectionChangedEventArgs e)
    {
      if (CustomersList.SelectedItem != null)
      {
        NavigationService.Navigate(new Uri("/Views/CustomerProjectsView.xaml?ID="
            + ((CustomerViewModel)CustomersList.SelectedItem).Id, UriKind.Relative));

        CustomersList.SelectedItem = null;
      }
    }

    private async void ResetDataMenuItem_Click(object sender, EventArgs e)
    {
      ApplicationBar.IsMenuEnabled = false;
      await ResetDataAsync();
      LoadData();
      ApplicationBar.IsMenuEnabled = true;
    }

    private async Task ResetDataAsync()
    {
      // Empty the Customer and Project tables 
      string sql = @"DELETE FROM Project";
      var db = await App.GetDatabaseAsync();

      await db.ExecuteStatementAsync(sql);
      sql = @"DELETE FROM Customer";
      await db.ExecuteStatementAsync(sql);

      CustomersList.ItemsSource = null;

      List<Task> tasks = new List<Task>();

      // Add seed customers and projects
      var cust1Task = InsertCustomer(db, "Adventure Works", "Bellevue", "Mu Han");
      tasks.Add(cust1Task.ContinueWith((id) => InsertProject(db, id.Result, "Expense Reports", "Windows Store app", DateTime.Today.AddDays(4))));
      tasks.Add(cust1Task.ContinueWith((id) => InsertProject(db, id.Result, "Time Reporting", "Windows Store app", DateTime.Today.AddDays(14))));
      tasks.Add(cust1Task.ContinueWith((id) => InsertProject(db, id.Result, "Project Management", "Windows Store app", DateTime.Today.AddDays(24))));

      var cust2Task = InsertCustomer(db, "Contoso", "Seattle", "David Hamilton");
      tasks.Add(cust2Task.ContinueWith((id) => InsertProject(db, id.Result, "Soccer Scheduling", "Windows Phone app", DateTime.Today.AddDays(6))));

      var cust3Task = InsertCustomer(db, "Fabrikam", "Redmond", "Guido Pica");
      tasks.Add(cust3Task.ContinueWith((id) => InsertProject(db, id.Result, "Product Catalog", "MVC4 app", DateTime.Today.AddDays(4))));
      tasks.Add(cust3Task.ContinueWith((id) => InsertProject(db, id.Result, "Expense Reports", "Windows Store app", DateTime.Today.AddDays(-3))));
      tasks.Add(cust3Task.ContinueWith((id) => InsertProject(db, id.Result, "Expense Reports", "Windows Phone app", DateTime.Today.AddDays(45))));

      var cust4Task = InsertCustomer(db, "Tailspin Toys", "Kent", "Michelle Alexander");
      tasks.Add(cust4Task.ContinueWith((id) => InsertProject(db, id.Result, "Kids Game", "Windows Store app", DateTime.Today.AddDays(60))));

      await Task.WhenAll(tasks.ToArray());
    }

    private async Task<int> InsertCustomer(SQLiteWinRT.Database db, string customerName, string customerCity, string customerContact)
    {
      try
      {
        using (var custstmt = await db.PrepareStatementAsync("INSERT INTO Customer (Name, City, Contact) VALUES (?, ?, ?)"))
        {
          custstmt.BindTextParameterAt(1, customerName);
          custstmt.BindTextParameterAt(2, customerCity);
          custstmt.BindTextParameterAt(3, customerContact);
          // To execute a prepared SQL statement, you call the StepAsync method which returns 
          // true if there are more rows to return, false if there are none - even if the statement does not return any rows. 
          // Confusion can arise if the developer thinks that it returns true or false according to the success or 
          // failure of the operation (in actual fact, failure is indicated by a thrown exception)
          await custstmt.StepAsync();

          Debug.WriteLine("INSERT completed OK for customer " + customerName);
        }
      }
      catch (Exception ex)
      {
        Debug.WriteLine(ex.Message);
        return 0;
      }

      using (var idstmt = await db.PrepareStatementAsync("SELECT ROWID FROM Customer WHERE Name = ? AND Contact = ?"))
      {
        idstmt.BindTextParameterAt(1, customerName);
        idstmt.BindTextParameterAt(2, customerContact);
        if (await idstmt.StepAsync())
        {
          return (int)idstmt.GetIntAt(0);
        }
        throw new Exception("Couldn't get inserted ID");
      };
    }

    private async Task InsertProject(SQLiteWinRT.Database db, int customerID, string name, string description, DateTime duedate)
    {
      using (var projstmt = await db.PrepareStatementAsync("INSERT INTO Project (CustomerId, Name, Description, DueDate) VALUES (?, ?, ?, ?)"))
      {

        // Reset the prepared statement so we can reuse it.
        projstmt.ClearBindings();
        projstmt.Reset();

        projstmt.BindIntParameterAt(1, customerID);
        projstmt.BindTextParameterAt(2, name);
        projstmt.BindTextParameterAt(3, description);
        projstmt.BindTextParameterAt(4, duedate.ToString("yyyy-MM-dd HH:mm:ss"));

        await projstmt.StepAsync();
        Debug.WriteLine("INSERT Project completed OK for customer " + customerID);
      }
    }
  }
}